import time
import traceback
from typing import List

from ._model import Model

from helpers.core.utils import Styler, APIException, get_version, print_exc, print_verbose, run_verbose, sprint
from helpers.sourcemanager import SourceManager
from helpers.options import BatchOptions

__version__ = get_version()


def _pause(sec):
    print_verbose('Pausing for 3 seconds...')
    time.sleep(sec)
    print_verbose('Waking up!')


def batch_download(source_strings: List[str], rootdir: str, batchOptions: BatchOptions):
    """Batch downloads model from CivitAI one by one."""

    source_manager = SourceManager()

    for id in source_manager.parse_src(source_strings):

        iter = 0
        while True:
            try:
                model = Model(id=id, dst_root_path=rootdir,
                              batchOptions=batchOptions).download()
                _pause(batchOptions.pause_time)
                break
            except Exception as e:
                sprint('---------')
                run_verbose(traceback.print_exc)
                print_exc(e, '\n')
                sprint('---------')
                _pause(batchOptions.pause_time)
                # if not isinstance(e, APIException):
                #     break
                if iter < batchOptions.retry_count:
                    sprint(Styler.stylize(
                        'Retrying to download the current model...', color='info'))
                    iter += 1
                else:
                    sprint(Styler.stylize(
                        f'Max retry of {batchOptions.retry_count} reached. Skipping the current model...', color='info'))
                    break



import os
import re
from typing import Dict, List, Union

import argparse

from helpers.core.utils import get_version
from helpers.argparse import PwdAction, ConfirmAction, ColoredArgParser, BooleanOptionalAction

__all__ = ['get_args']


def add_shared_option(par):
    par.add_argument(
        '--with-color', action=BooleanOptionalAction, help='Enable styles like colors, background colors and bold/italized texts.'
    )
    par.add_argument(
        '--verbose', action=BooleanOptionalAction, help='Prints out traceback and other useful information.')
    par.add_argument(
        '-v', '--version', action='version', version=f'civitdl v{get_version()}', help='Prints out the version of the program.'
    )


parser = ColoredArgParser(
    prog='civitconfig',
    description="civitconfig is a cli program used to set configurations for the main program, civitdl.",
    formatter_class=argparse.RawTextHelpFormatter
)


subparsers = parser.add_subparsers(
    dest='subcommand',
    required=True,
    help='Choose one of the following subcommands: default, sorter, alias.')

default_parser = subparsers.add_parser(
    'default', help='Set a default value for one of the options below.\nIf no options are provided, default will print the current default.')

default_parser.add_argument('-i', '--max-images', metavar='INT', type=int,
                            help='Set the default max number of images to download per model.')

default_parser.add_argument('--nsfw-mode', type=str,
                            help='Set the default nsfw mode when downloading images. Setting to 0 means the program will always download sfw. Setting to 1 means the program will only download sfw + nsfw images if the model itself is nsfw. Setting to 2 means the program will always download both sfw and nsfw.')

default_parser.add_argument('-s', '--sorter', metavar='NAME', type=str,
                            help='Set the default sorter given name of sorter (filepath not accepted).')

default_parser.add_argument('-k', '--api-key', action=PwdAction, type=str, required=False, nargs='?',
                            help='Prompts the user for their api key to use for model downloads that require users to log in.')

default_parser.add_argument('--with-prompt', action=BooleanOptionalAction,
                            help='Toggles default behavior on whether to download prompt alongside images.')

default_parser.add_argument('--without-model', action=BooleanOptionalAction,
                            help='Toggles default behavior on whether to download model or not.')

default_parser.add_argument('--limit-rate', type=str,
                            help='Set the default limit for the download speed/rate of resources downloaded from CivitAI. Set it to 0 to disable limit.'
                            )

default_parser.add_argument('--retry-count', type=int,
                            help='Set the default max number of times to retry downloading a model if it fails.'
                            )

default_parser.add_argument('--pause-time', type=float,
                            help='Set the default number of seconds to pause between each model\'s download'
                            )
default_parser.add_argument('--cache-mode', type=str,
                            help='Set the default cache mode. Valid modes are 0 and 1 for now. 2 is not implemented yet.')
default_parser.add_argument('--model-overwrite', action=BooleanOptionalAction,
                            help='Set the default behavior on whether to overwrite or skip models that are already downloaded at path. model=overwrite to overwrite model. no-model-overwrite to skip model.')
add_shared_option(default_parser)


sorter_parser = subparsers.add_parser(
    'sorter', help='Sorter-related subcommand.\nCurrently supports listing, adding and deleting sorters.\nIf no options are provided, sorter will list all available sorters.\nExample usage: civitdl 123456 ./lora --sorter mysorter.',
    formatter_class=argparse.RawTextHelpFormatter
)
sorter_group = sorter_parser.add_mutually_exclusive_group()
sorter_group.add_argument('-a', '--add', metavar=('NAME', 'FILEPATH'), type=str, nargs=2,
                          help='Add/save a new sorter to civitdl program.\nExample: civitconfig sorter --add mysorter ./custom/tags.py.')
sorter_group.add_argument('-d', '--delete', metavar='NAME', type=str,
                          help='Delete sorter based on name of the sorter.\nExample: civitconfig sorter --delete mysorter.')
add_shared_option(sorter_parser)


alias_parser = subparsers.add_parser(
    'alias',
    help='Alias-related subcommand.\nCurrently supports listing, adding and deleting aliases.\nIf no options are provided, alias will list all available aliases.\nExample usage: civitdl 123456 @lora.',
    formatter_class=argparse.RawTextHelpFormatter
)
alias_group = alias_parser.add_mutually_exclusive_group()
alias_group.add_argument('-a', '--add', metavar=('NAME', 'FILEPATH'), type=str, nargs=2,
                         help='Add a new alias to the civitdl program.\nExample: civitconfig alias --add @lora ./ComfyUI/models/loras.')
alias_group.add_argument('-d', '--delete', metavar=('NAME'), type=str,
                         help='Delete alias based on name of the alias.\nExample: civitconfig alias --delete @lora.')
add_shared_option(alias_parser)


config_parser = subparsers.add_parser(
    'settings',
    help='Subcommand related to the operations of civitconfig.',
    formatter_class=argparse.RawTextHelpFormatter
)
config_group = config_parser.add_mutually_exclusive_group()
config_group.add_argument(
    '-r', '--reset', action=ConfirmAction, help='Delete config and reinstall the default configuration.\nWARNING: DO NOT RUN THIS UNLESS YOU ARE SURE YOU WANT TO DELETE.')
config_group.add_argument(
    '-d', '--download', metavar=('PATH'), type=str, help='Downloads config directory to specified path.')
add_shared_option(config_parser)


def get_args():
    parser_result = parser.parse_args()

    return vars(parser_result)



import os
import shutil
import sys
import time
import csv
from typing import IO, Callable, Iterable, Union

from ._ui.styler import Styler, InputException, UnexpectedException
from .utils import get_progress_bar, sprint


class IOHelper:

    # Level 0 #

    @staticmethod
    def createDirsIfNotExist(dirpaths):
        for dirpath in dirpaths:
            os.makedirs(dirpath, exist_ok=True)

    @staticmethod
    def delete_file_if_exists(file_path):
        if os.path.exists(file_path):
            try:
                os.remove(file_path)
            except OSError as e:
                raise UnexpectedException(
                    f"Error deleting file {file_path}: {e}")

    @staticmethod
    def write_contents(file: IO, content_chunks: Iterable, limit_rate: Union[int, None] = None, update_pb: Union[Callable[[int], None], None] = None):
        last_chunk_time = time.perf_counter()
        for content in content_chunks:
            file.write(content)
            bytes_downloaded = len(content)

            download_time = time.perf_counter() - last_chunk_time
            speed = bytes_downloaded / \
                download_time if download_time != 0 else float('inf')
            if limit_rate is not None and limit_rate != 0 and speed > limit_rate:
                time_to_sleep = (bytes_downloaded / limit_rate) - download_time
                if time_to_sleep > 0:
                    time.sleep(time_to_sleep)

            if update_pb:
                update_pb(bytes_downloaded)
            last_chunk_time = time.perf_counter()

    @staticmethod
    def read_dict_from_csv(filepath: str):
        csv_dict = {}

        with open(filepath, 'r', encoding='UTF-8') as f:
            csv_iterator = csv.reader(f)
            next(csv_iterator)
            for row in csv_iterator:
                try:
                    row[2]
                    raise InputException(
                        f'CSV file is invalid. File contains rows with more than two columns.\nThe filepath for the invalid csv is "{filepath}"')
                except:
                    csv_dict[row[0]] = row[1]

        return csv_dict

    # Level 1 #
    @classmethod
    def write_to_file(cls, filepath: str, content_chunks: Iterable, mode: str = None, limit_rate: Union[int, None] = 0, encoding: Union[str, None] = None, overwrite: bool = True, use_pb: bool = False, total: float = 0, desc: str = None):
        """Uses content_chunks to write to filepath bit by bit. If use_pb is enabled, it is recommended to set total kwarg to the length of the file to be written."""
        progress_bar = get_progress_bar(total, desc) if use_pb else None

        def update_progress_bar(bytes_downloaded):
            if progress_bar:
                progress_bar.update(bytes_downloaded)

        if not overwrite and os.path.exists(filepath):
            if (progress_bar):
                progress_bar.close()
            sprint(Styler.stylize(
                f'File already exists at "{filepath}"', color='info'))
        else:
            temp_dirpath = os.path.join(os.path.dirname(
                filepath), '.tmp')
            temp_filepath = os.path.join(
                temp_dirpath, os.path.basename(filepath))
            os.makedirs(temp_dirpath, exist_ok=True)
            try:
                with open(temp_filepath, mode if mode != None else 'w', encoding=encoding) as file:
                    cls.write_contents(file, content_chunks,
                                       limit_rate, update_progress_bar)
                shutil.move(temp_filepath, filepath)
                shutil.rmtree(temp_dirpath)
            except Exception as e:
                sprint('Existance: ', temp_dirpath, temp_filepath,
                       os.path.exists(temp_filepath), file=sys.stderr)
                if os.path.exists(temp_dirpath):
                    shutil.rmtree(temp_dirpath)
                raise e
            if (progress_bar):
                progress_bar.close()

    @classmethod
    def write_to_files(cls, dirpath: str, basenames: Iterable, content_chunks_list: Iterable[Iterable], mode: str = None, encoding: Union[str, None] = None, use_pb: bool = False, total: float = 0, desc: str = None):
        """Write content to multiple files in dirpath. If use_pb is enabled, it is recommended to set total kwarg to the number of files being written."""
        progress_bar = get_progress_bar(total, desc) if use_pb else None
        for basename, content_chunks in zip(basenames, content_chunks_list):
            filepath = os.path.join(dirpath, basename)
            with open(filepath, mode if mode != None else 'w', encoding=encoding) as file:
                cls.write_contents(file, content_chunks)
                if (progress_bar):
                    progress_bar.update(1)
        if (progress_bar):
            progress_bar.close()



from typing import Callable, Dict, List, Literal, Union, Optional
import requests

from helpers.sorter.utils import SorterData, import_sort_model
from helpers.sorter import basic, tags
from helpers.core.utils import disable_style, InputException, NotImplementedException, UnexpectedException, Validation, print_verbose, safe_run, set_verbose


def parse_bytes(size: Union[str, int, float], name: str):
    units = {"k": 10**3, "m": 10**6, "g": 10**9, "t": 10**12}

    res = safe_run(float, size)
    if res["success"] == True:
        number = round(res["data"])
        Validation.validate_integer(number, name)
        return number
    else:
        number, unit = (size[0:-1], size[-1:].lower())

        num_res = safe_run(float, number)
        if num_res["success"] == False:
            raise InputException(
                f'Invalid byte value for {name}: {size}')
        else:
            number = round(num_res["data"])
            Validation.validate_integer(number, name)

        if unit not in units:
            raise InputException(
                f'Invalid byte value for {name}: {unit}')
        else:
            return number * units[unit]


class BatchOptions:
    sorter_name: str
    sorter: Callable[[Dict, Dict, str, str],
                     SorterData] = basic.sort_model
    max_images: int = 3
    nsfw_mode: Literal['0', '1', '2'] = '1'
    api_key: Optional[str] = None

    with_prompt: bool = True
    without_model: bool = False
    limit_rate: int = 0
    retry_count: int = 3
    pause_time: int = 3

    cache_mode: Literal['0', '1', '2'] = '1'
    model_overwrite: bool = False

    with_color: bool = False
    verbose: Optional[bool] = None

    def __get_sorter(self, sorter: str):
        if not isinstance(sorter, property) and not isinstance(sorter, str):
            raise InputException(
                'Sorter provided is not a string in BatchOptions.')

        sorter_str = 'basic' if isinstance(sorter, property) else sorter
        if sorter_str == 'basic' or sorter_str == 'tags':
            self._sorter = tags.sort_model if sorter_str == 'tags' else basic.sort_model
        else:
            self._sorter = import_sort_model(sorter_str)

        print_verbose("Chosen Sorter Description: ", self._sorter.__doc__)
        return self._sorter

    def __init__(self, retry_count, pause_time, max_images, nsfw_mode, with_prompt, without_model, api_key, verbose, sorter, limit_rate, cache_mode, model_overwrite, with_color):
        self.session = requests.Session()

        # FIXME: Move usage of with_color and verbose outside of options
        if with_color is not None:
            Validation.validate_bool(with_color, 'with_color')
            self.with_color = with_color
            if with_color == False:
                disable_style()

        if verbose is not None:
            Validation.validate_bool(verbose, 'verbose')
            self.verbose = verbose
            if verbose == True:
                set_verbose(self.verbose)

        if sorter is not None:
            Validation.validate_string(sorter, 'sorter')
            self.sorter = self.__get_sorter(sorter)
            self.sorter_name = sorter

        if max_images is not None:
            Validation.validate_integer(max_images, 'max_images', min_value=0)
            self.max_images = max_images

        if nsfw_mode is not None:
            Validation.validate_string(
                nsfw_mode, 'nsfw_mode', whitelist=['0', '1', '2'])
            self.nsfw_mode = nsfw_mode

        if api_key is not None and api_key != '':
            Validation.validate_string(api_key, 'api_key')
            self.api_key = api_key

        if with_prompt is not None:
            Validation.validate_bool(with_prompt, 'with_prompt')
            self.with_prompt = with_prompt

        if without_model is not None:
            Validation.validate_bool(without_model, 'without_model')
            self.without_model = without_model

        if limit_rate is not None:
            Validation.validate_types(
                limit_rate, [str, int, float], 'limit_rate')
            self.limit_rate = parse_bytes(limit_rate, "limit_rate")

        if retry_count is not None:
            Validation.validate_integer(
                retry_count, 'retry_count', min_value=0)
            self.retry_count = retry_count

        if pause_time is not None:
            Validation.validate_float(pause_time, 'pause_time', min_value=0)
            self.pause_time = pause_time

        if cache_mode is not None:
            Validation.validate_string(
                cache_mode, 'cache_mode', whitelist=['0', '1', '2'])
            if cache_mode == '2':
                raise NotImplementedException(
                    'cache mode of 2 has not been implemented yet')
            self.cache_mode = cache_mode

        if model_overwrite is not None:
            Validation.validate_bool(model_overwrite, 'model_overwrite')
            self.model_overwrite = model_overwrite


class DefaultOptions:
    sorter: Optional[str] = None
    max_images: Optional[int] = None
    nsfw_mode: Optional[int] = None
    api_key: Optional[str] = None

    with_prompt: Optional[bool] = None
    without_model: Optional[bool] = None
    limit_rate: Optional[str] = None
    retry_count: Optional[int] = None
    pause_time: Optional[int] = None

    cache_mode: Optional[int] = None
    model_overwrite: Optional[bool] = None

    with_color: Optional[bool] = None

    def __init__(self, sorter=None, max_images=None, nsfw_mode=None, api_key=None, with_prompt=None, without_model=None, limit_rate=None, retry_count=None, pause_time=None, cache_mode=None, model_overwrite=None, with_color=None):
        if sorter is not None:
            Validation.validate_string(
                sorter, 'sorter')

            self.sorter = sorter

        if max_images is not None:
            Validation.validate_integer(
                max_images, 'max_images', min_value=0
            )
            self.max_images = max_images

        if nsfw_mode is not None:
            Validation.validate_string(
                nsfw_mode, 'nsfw_mode', whitelist=['0', '1', '2']
            )
            self.nsfw_mode = nsfw_mode

        if api_key is not None:
            Validation.validate_string(api_key, 'api_key')
            self.api_key = api_key

        if with_prompt is not None:
            Validation.validate_bool(with_prompt, 'with_prompt')
            self.with_prompt = with_prompt

        if without_model is not None:
            Validation.validate_bool(without_model, 'without_model')
            self.without_model = without_model

        if limit_rate is not None:
            Validation.validate_string(limit_rate, 'limit_rate')
            parse_bytes(limit_rate, 'limit_rate')
            self.limit_rate = limit_rate

        if retry_count is not None:
            Validation.validate_integer(
                retry_count, 'retry_count', min_value=0
            )
            self.retry_count = retry_count

        if pause_time is not None:
            Validation.validate_float(
                pause_time, 'pause_time', min_value=0
            )
            self.pause_time = pause_time

        if cache_mode is not None:
            Validation.validate_string(
                cache_mode, 'cache_mode', whitelist=['0', '1', '2']
            )
            if cache_mode == '2':
                raise NotImplementedException(
                    'cache mode of 2 has not been implemented yet')
            self.cache_mode = cache_mode

        if model_overwrite is not None:
            Validation.validate_bool(model_overwrite, 'model_overwrite')
            self.model_overwrite = model_overwrite

        if with_color is not None:
            Validation.validate_bool(with_color, 'with_color')
            self.with_color = with_color




import argparse
from gettext import gettext
import os
import sys
import getpass
import warnings
try:
    import msvcrt
except:
    None

from helpers.core.utils import Styler


def windows_getpass(prompt=""):
    api_key = ""
    while True:
        char = msvcrt.getch().decode('utf-8')
        if char == '\r' or char == '\n':
            break
        elif char == '\x03':  # Ctrl+C
            raise KeyboardInterrupt
        elif char == '\b':
            api_key = api_key[:-1]
        else:
            api_key += char
    return api_key


class PwdAction(argparse.Action):

    def __call__(self, parser, namespace, values, option_string=None):
        mypass = values
        if mypass is None:
            mypass = windows_getpass(prompt='Key: ') if os.name == 'nt' else getpass.getpass(
                prompt='Key: ')
        setattr(namespace, self.dest, mypass)


class ColoredArgParser(argparse.ArgumentParser):

    def print_usage(self, file=None):
        if file is None:
            file = sys.stdout
        self._print_message(self.format_usage()[0].upper() +
                            self.format_usage()[1:],
                            file, 'warning')

    def print_help(self, file=None):
        if file is None:
            file = sys.stdout
        self._print_message(self.format_help()[0].upper() +
                            self.format_help()[1:],
                            file, 'info')

    def _print_message(self, message: str, file=None, color=None, styles=None):
        if message:
            if file is None:
                file = sys.stderr
            if color is None:
                file.write(message)
            else:
                file.write(
                    Styler.stylize(f'{message.strip()}\n', color=color, styles=styles))

    def exit(self, status=0, message=None):
        if message:
            self._print_message(message, sys.stderr,
                                'exception')
        sys.exit(status)

    def error(self, message):
        self.print_usage(sys.stderr)
        args = {'prog': self.prog, 'message': message}
        self.exit(2, gettext('%(prog)s: Error: %(message)s\n') % args)


# Copied from argparse for compatibility reasons below python v3.9
_deprecated_default = object()


class BooleanOptionalAction(argparse.Action):
    def __init__(self,
                 option_strings,
                 dest,
                 default=None,
                 type=_deprecated_default,
                 choices=_deprecated_default,
                 required=False,
                 help=None,
                 metavar=_deprecated_default):

        _option_strings = []
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith('--'):
                option_string = '--no-' + option_string[2:]
                _option_strings.append(option_string)

        # We need `_deprecated` special value to ban explicit arguments that
        # match default value. Like:
        #   parser.add_argument('-f', action=BooleanOptionalAction, type=int)
        for field_name in ('type', 'choices', 'metavar'):
            if locals()[field_name] is not _deprecated_default:
                warnings._deprecated(
                    field_name,
                    "{name!r} is deprecated as of Python 3.12 and will be "
                    "removed in Python {remove}.",
                    remove=(3, 14))

        if type is _deprecated_default:
            type = None
        if choices is _deprecated_default:
            choices = None
        if metavar is _deprecated_default:
            metavar = None

        super().__init__(
            option_strings=_option_strings,
            dest=dest,
            nargs=0,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            setattr(namespace, self.dest, not option_string.startswith('--no-'))

    def format_usage(self):
        return ' | '.join(self.option_strings)


class ConfirmAction(BooleanOptionalAction):
    def __call__(self, parser, namespace, value, option_string=None):
        response = input(
            'Are you sure you want to move the current configuration to trash, and reset with default? (y/N): ').strip().lower()
        if response not in ('y', 'yes'):
            parser.error(f"Reset declined. Exiting.")

        super().__call__(parser, namespace, value, option_string)



from helpers.core.utils import Styler, InputException, ResourcesException, UnexpectedException, APIException, sprint, print_verbose

from helpers.sourcemanager import Id
from helpers.options import BatchOptions


class Metadata:
    __batchOptions = None
    __id = None

    model_name = None
    model_id = None
    version_id = None
    version_hashes = None

    model_dict = None
    version_dict = None

    nsfw = False
    nsfwLevel = None
    image_dicts = None
    image_download_urls = None
    model_download_url = None

    def __init__(self, id: Id, batchOptions: BatchOptions):
        self.__id = id
        self.__batchOptions = batchOptions

    def make_api_call(self):
        type = self.__id.type
        ids = self.__id.data

        if type == 'site' and len(ids) == 2:
            self.model_id = ids[0]
            self.version_id = ids[1]
            self.model_dict = self.__get_model_metadata()
            self.version_dict = self.__get_version_metadata()
        elif type == 'id' or (type == 'site' and len(ids) == 1):
            self.model_id = ids[0]
            self.model_dict = self.__get_model_metadata()
            if len(self.model_dict['modelVersions']) == 0:
                raise ResourcesException(
                    f'No model versions found from model id, {self.model_id}')
            self.version_dict = self.model_dict['modelVersions'][0]
            self.version_id = str(self.version_dict['id'])
        elif type == 'api':
            self.version_id = ids[0]
            self.version_dict = self.__get_version_metadata()
            self.model_id = str(self.version_dict['modelId'])
            self.model_dict = self.__get_model_metadata()
        else:
            raise InputException(
                f'Incorrect format sent ({ids}, {type}): "{self.__id.original}"')

        self.model_name = self.model_dict['name']

        self.nsfw = self.model_dict['nsfw']
        self.nsfwLevel = self.model_dict['nsfwLevel']

        print_verbose(f'nsfwLevel of model: {self.nsfwLevel}')
        print_verbose(f'Configurated nsfw mode: {self.__batchOptions.nsfw_mode}')  # nopep8

        self.image_dicts = []
        for image_dict in self.version_dict['images']:
            if image_dict['nsfwLevel'] <= 3:
                self.image_dicts.append(image_dict)
            elif (self.__batchOptions.nsfw_mode == '1' and (self.nsfw or image_dict['nsfwLevel'] <= self.nsfwLevel)) or self.__batchOptions.nsfw_mode == '2':
                self.image_dicts.append(image_dict)
        self.image_dicts = self.image_dicts[0:self.__batchOptions.max_images]

        print_verbose(f'nsfwLevel of images to download: {[image_dict['nsfwLevel'] for image_dict in self.image_dicts]}')  # nopep8

        self.image_download_urls = [image_dict['url']
                                    for image_dict in self.image_dicts]

        self.model_download_url = self.version_dict['downloadUrl']

        self.version_hashes = self.__get_version_hashes(
            self.version_dict, self.model_download_url)

        return self

    def __get_version_hashes(self, version_dict, download_url):
        files = version_dict['files']
        hashes = {}
        for file in files:
            if "downloadUrl" in file and file['downloadUrl'] == download_url:
                if isinstance(file['hashes'], dict):
                    hashes = file['hashes']
                else:
                    sprint(Styler.stylize(
                        'Hashes found in metadata is not a dictionary! There is an error with the API!', color='error'))
                break
        if hashes is {}:
            sprint(Styler.stylize('Hash not found in metadata.', color='warning'))
        elif "SHA256" not in hashes:
            sprint(Styler.stylize('SHA256 hash not found.', color='warning'))
        return hashes

    def __get_model_metadata(self):
        """Returns json object if request succeeds, else print error and returns None"""
        metadata_url = f'https://civitai.com/api/v1/models/{self.model_id}'
        metadata = self.__get_metadata(metadata_url)
        return metadata

    def __get_version_metadata(self):
        metadata_url = f'https://civitai.com/api/v1/model-versions/{self.version_id}'  # nopep8 # Linux issue
        metadata = self.__get_metadata(metadata_url)
        return metadata

    def __get_metadata(self, url: str):
        print_verbose('Requesting model metadata.')
        print_verbose(f'Metadata API Request URL: {url}')
        meta_res = self.__batchOptions.session.get(url, stream=True)

        print_verbose('Finished requesting model metadata.')
        if meta_res.status_code != 200:
            raise APIException(
                meta_res.status_code, f'Downloading metadata from CivitAI for "{self.__id.original}" failed when trying to request metadata from "{url}"')

        try:
            return meta_res.json()
        except Exception as e:
            raise UnexpectedException(
                'Unable to parse metadata from CivitAI (incorrect format provided by Civitai).', 'CivitAI might be under maintainence.',
                f'\nOriginal Error:\n       {e}')



from json import dumps, loads
import shutil
import os
import re
from typing import Dict, List, Union
from math import ceil

import requests

from helpers.core.utils import Styler, InputException, UnexpectedException, APIException, print_newlines, sprint, print_verbose, concurrent_request
from helpers.core.iohelper import IOHelper

from helpers.sourcemanager import Id
from helpers.options import BatchOptions
from helpers.hashmanager import HashManager

from ._metadata import Metadata


class Model:
    __id: Id
    __dst_root_path: str
    __batchOptions: BatchOptions

    def __init__(self, id: Id, dst_root_path: str, batchOptions: BatchOptions):
        self.__id = id
        self.__dst_root_path = dst_root_path
        self.__batchOptions = batchOptions

    def __download_images(self, dirpath: str, urls: List[str], filenames: List[str]):
        def make_req(url): return self.__batchOptions.session.get(
            url, stream=True)

        # TODO: Change progress bar to be based on time length of request rather than when all the images are fetched and ready to be written.
        # TODO: what if a specific image have a hard time with getting a response?

        os.makedirs(dirpath, exist_ok=True)
        print_verbose('Now requesting images...')
        image_content_chunks_list = [res.iter_content(1024*1024) for res in concurrent_request(
            req_fn=make_req, urls=urls)]
        print_verbose('Finished requesting images...')

        if (len(image_content_chunks_list) == 0):
            sprint(Styler.stylize('No images to download...', color='warning'))
        else:
            IOHelper.write_to_files(dirpath, filenames, image_content_chunks_list, mode='wb',
                                    use_pb=True, total=len(filenames), desc='Images')

    def __download_prompt(self, dirpath: str, filenames: List[str], prompts: List[Dict]):
        if len(prompts) != 0:
            os.makedirs(dirpath, exist_ok=True)
            IOHelper.write_to_files(dirpath, filenames, [dumps(
                image_dict, indent=2, ensure_ascii=False) for image_dict in prompts], encoding='UTF-8')

    def __download_metadata(self, dirpath: str, filename: str, metadata: Metadata):
        os.makedirs(dirpath, exist_ok=True)
        filepath = os.path.join(dirpath, filename)
        IOHelper.write_to_file(filepath, [dumps(
            metadata.model_dict, indent=2, ensure_ascii=False)], encoding='UTF-8')

    def __download_hash(self, dirpath: str, filename: str, hashes: Dict):
        os.makedirs(dirpath, exist_ok=True)
        filepath = os.path.join(dirpath, filename)
        data = 'hash_name, hash_id\n'
        for key, value in hashes.items():
            data += f'{key}, {value}\n'
        IOHelper.write_to_file(
            filepath, [data.rstrip()], encoding='UTF-8')

    def __download_model(self, dirpath, filename: str, model_res: requests.Response, metadata: Metadata):
        os.makedirs(dirpath, exist_ok=True)
        filepath = os.path.join(dirpath, filename)
        hash_manager = HashManager(
            metadata.version_id) if self.__batchOptions.cache_mode != '0' else None
        cached_filepath = hash_manager.get_local_model_path() if hash_manager else None

        if cached_filepath and cached_filepath != os.path.abspath(filepath):
            print_newlines(Styler.stylize(f"""Model already existed at the following path:
                - Path: {cached_filepath}""", color='info'))
            sprint(Styler.stylize(f"Copying to new path...", color='info'))
            shutil.copy(cached_filepath, filepath)
        else:
            content_chunks = model_res.iter_content(
                ceil(self.__batchOptions.limit_rate / 8)
                if self.__batchOptions.limit_rate is not None and self.__batchOptions.limit_rate != 0
                else 1024*1024)
            IOHelper.write_to_file(filepath, content_chunks, mode='wb', limit_rate=self.__batchOptions.limit_rate,
                                   overwrite=self.__batchOptions.model_overwrite, use_pb=True, total=float(model_res.headers.get('content-length', 0)), desc='Model')
        if hash_manager:
            hash_manager.set_local_model_cache(
                filepath, metadata.version_hashes)

    def __request_model(self, metadata: Metadata):
        # Request model
        print_verbose('Preparing to send download model request...')
        print_verbose(f'Model Download API URL: {metadata.model_download_url}')

        headers = {
            'Authorization': f'Bearer {self.__batchOptions.api_key}',
        }
        res = self.__batchOptions.session.get(
            metadata.model_download_url, stream=True, headers=headers)

        print_verbose(f'Status Code: {res.status_code}')

        if 'reason=download-auth' in res.url or res.status_code == 401:
            print_verbose('Unauthorized status', res.url,  res.status_code)
            raise InputException(
                'Unable to download this model as it requires a valid API Key. Please head to "civitai.com", go to "Account Settings", then go to "API Keys" section, then add an api key to your account. After that, paste the key to the program or add it to civitconfig with "civitconfig default --api-key".')

        if res.status_code == 403:
            sprint(Styler.stylize(
                'Model is behind CivitAI\'s early access restriction (i.e have to wait a few days for model to be available to be downloaded).', color='warning'))
            try:
                data = loads(res.content)
                sprint(Styler.stylize(f'Deadline: {data["deadline"]}', color='warning'))  # nopep8
                sprint(Styler.stylize(f'Error Message: {data["message"]}', color='warning'))  # nopep8
            except:
                None
        if res.status_code != 200:
            raise APIException(
                res.status_code, f'Downloading model from CivitAI failed for model id, {metadata.model_id}, and version id, {metadata.version_id}')

        return res

    def __get_filenames(self, metadata: Metadata, content_disposition: Union[str, None] = None):
        def get_api_filename():
            version_file = None
            for file in metadata.version_dict['files']:
                file_version_id = re.search(
                    r'(?<=models\/)\d+', file['downloadUrl'])
                if file_version_id != None and file_version_id.group(0) == metadata.version_id:
                    version_file = file
            filename = None

            if content_disposition == None:
                sprint(Styler.stylize(
                    f'Downloaded model from CivitAI has no content disposition header available.', color='warning'))
                filename = f'{metadata.model_name}--{version_file["name"]}'
            else:
                try:
                    filename = content_disposition.split(
                        'filename=')[-1].strip('"').encode('latin-1').decode('utf-8')
                except UnicodeDecodeError as e:
                    # Alternative solution for finding filename
                    sprint(Styler.stylize(e, color='warning'))
                    filename = f'{metadata.model_name}--{version_file["name"]}'

            if filename == None:
                raise UnexpectedException(
                    f'Unable to retrieve filename for model with model id, {metadata.model_id}, and version id, {metadata.version_id}')

            return filename

        # get filename of metadata
        metadata_filename = f'model_dict-mid_{metadata.model_id}-vid_{metadata.version_id}.json'  # nopep8
        # get filename of images
        image_filenames = [os.path.basename(url)
                           for url in metadata.image_download_urls]
        # get filename of prompts
        prompt_filenames = [
            f'{os.path.splitext(basename)[0]}.json' for basename in image_filenames]
        # get model_stem and model_ext
        model_stem, model_ext = os.path.splitext(get_api_filename())

        # get filename of model
        model_filename = f'{model_stem}-mid_{metadata.model_id}-vid_{metadata.version_id}{model_ext}'  # nopep8

        # get filename of hash
        hash_filename = f'{model_stem}-mid_{metadata.model_id}-vid_{metadata.version_id}.csv'  # nopep8

        return {
            'metadata': metadata_filename,
            'images': image_filenames,
            'prompts': prompt_filenames,
            'model': model_filename,
            'hash': hash_filename
        }

    def download(self):
        # 1. Get metadata
        metadata = Metadata(self.__id, self.__batchOptions).make_api_call()

        print_newlines(Styler.stylize(
            f"""Now downloading \"{metadata.model_name}\"...
                - Model ID: {metadata.model_id}
                - Version ID: {metadata.version_id}\n""",
            color='main'))

        # 2. Get directory and file paths

        model_res = self.__request_model(
            metadata) if not self.__batchOptions.without_model else None

        filenames = self.__get_filenames(
            metadata, model_res.headers.get('Content-Disposition') if model_res else None)

        sorter_data = self.__batchOptions.sorter(
            metadata.model_dict, metadata.version_dict, os.path.split(filenames['model'])[0], self.__dst_root_path)

        # 3. Download model and etc.

        self.__download_metadata(
            sorter_data.metadata_dir_path, filenames['metadata'], metadata)
        self.__download_images(sorter_data.image_dir_path,
                               metadata.image_download_urls, filenames['images'])
        if self.__batchOptions.with_prompt:  # FIXME: I don't like how one of them is with_prompt, and the other is without_model
            self.__download_prompt(
                sorter_data.prompt_dir_path, filenames['prompts'], metadata.image_dicts)
        if not self.__batchOptions.without_model:
            self.__download_model(sorter_data.model_dir_path,
                                  filenames['model'], model_res, metadata)
        self.__download_hash(sorter_data.model_dir_path,
                             filenames['hash'], metadata.version_hashes)

        print_newlines(Styler.stylize(
            f"""\nDownload completed for \"{metadata.model_name}\"
                - Model ID: {metadata.model_id}
                - Version ID: {metadata.version_id}
                - Model Directory Path: {sorter_data.model_dir_path if not self.__batchOptions.without_model else 'N/A'}
                - Hashes Directory Path: {sorter_data.model_dir_path}
                - Metadata Directory Path: {sorter_data.metadata_dir_path}
                - Images Directory Path: {sorter_data.image_dir_path}
                - Images Metadata Directory Path: {sorter_data.prompt_dir_path if self.__batchOptions.with_prompt else 'N/A'}\n""", color='success'))
        sprint('---------------------------\n')

        return self



from datetime import datetime
import sys
import importlib.metadata
from typing import Callable
import concurrent.futures
from tqdm import tqdm

from ._ui.styler import Styler, disable_style, CustomException, InputException, ResourcesException, UnexpectedException, APIException, NotImplementedException
from ._validation import Validation

# Level 0

_verbose = False


def get_verbose():
    return _verbose


def set_verbose(verbose: bool):
    global _verbose
    _verbose = verbose
    return _verbose


def getDate():
    now = datetime.now()
    return now.strftime("%Y_%m_%d-%Hm_%Mm_%Ss_%f")


def get_version():
    return importlib.metadata.version('civitdl')


def sprint(*args, **kwargs):
    try:
        print(*args, **kwargs)
    except:
        encoded_args = [str(arg).encode(
            'utf-8', 'replace') for arg in args]
        print(*encoded_args, **kwargs)


# Level 1 - Currently or in the future might depends on level 0


def run_verbose(fn, *args, **kwargs):
    if get_verbose():
        fn(*args, **kwargs)


def print_verbose(*args, **kwargs):
    if get_verbose():
        args = [Styler.stylize(str(arg), bg_color='info') for arg in args]
        sprint(*args, **kwargs)


def print_exc(exc: Exception, *args, **kwargs):
    if isinstance(exc, CustomException):
        sprint(exc, file=sys.stderr, *args, **kwargs)
    else:
        sprint(Styler.stylize(str(exc), color='exception'), *args,
               file=sys.stderr, **kwargs)


def print_newlines(string: str, **kwargs):
    for el in string.split('\n'):
        sprint(el, **kwargs)


def print_error(*args, **kwargs):
    args = [Styler.stylize(str(arg), bg_color='error') for arg in args]
    sprint(*args, **kwargs, file=sys.stderr)


def safe_run(callback: Callable[..., any], *values: any) -> dict:
    try:
        return {"success": True, "data": callback(*values)}
    except Exception as e:
        return {"success": False, "error": e}


def concurrent_request(req_fn, urls, max_workers=16):
    res_list = None

    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(req_fn, url) for url in urls]
        res_list = [future.result() for future in futures]

    return res_list


def get_progress_bar(total: float, desc: str):
    return tqdm(total=total, desc=desc,
                unit='iB', unit_scale=True, file=sys.stdout)


def find_in_list(li, cond_fn: Callable[[any, int], bool], default=None):
    """Given a list and a condition function, where the first argument is the index of the item and the second argument is the value of the item, it returns the first value in the list when the condition is true"""
    return next((item for i, item in enumerate(li) if cond_fn(item, int)), default)




import os
import json
from datetime import datetime
import shutil
from typing import Dict, List, Union

from helpers.sorter import basic, tags
from helpers.core.utils import Styler, UnexpectedException, getDate, sprint
from helpers.core._validation import Validation

DEFAULT_CONFIG = {
    "version": "1",
    "default": {
        "sorter": "basic",
        "max_images": 3,
        "nsfw_mode": "2",
        "api_key": "",

        "with_prompt": True,
        "without_model": False,
        "limit_rate": '0',
        "retry_count": 3,
        "pause_time": 3.0,

        "cache_mode": '1',
        "model_overwrite": False,

        "with_color": True
    },
    "sorters": [["basic", basic.sort_model.__doc__, 'basic'], ["tags", tags.sort_model.__doc__, 'tags']],
    "aliases": [["@example", "~/.models"]]
}


CURRENT_VERSION = "1"

_config = None  # Do not do async operations with this


class Config:

    _config_path: str
    _config_dir_path: str
    _config_trash_dir_path: str
    _sorters_dir_path: str
    _sorters_trash_dir_path: str

    def __init__(self, config_path, config_dir_path, config_trash_dir_path, sorters_dir_path, sorters_trash_dir_path):
        self._config_path = config_path
        self._config_dir_path = config_dir_path
        self._config_trash_dir_path = config_trash_dir_path
        self._sorters_dir_path = sorters_dir_path
        self._sorters_trash_dir_path = sorters_trash_dir_path

    # Private Methods #

    ## Configs ##

    def _configExists(self):
        return os.path.exists(self._config_path)

    def _getConfig(self):
        global _config
        if not _config:
            with open(self._config_path, encoding='UTF-8') as file:
                _config = json.load(file)
            if _config == None:
                raise UnexpectedException('JSON config file was not read...')

        return _config

    def _saveConfig(self, dic: Dict):
        global _config
        with open(self._config_path, 'w', encoding='UTF-8') as file:
            json.dump(dic, file, indent=2)
            _config = dic

    def _get_valid_keys(self):
        return DEFAULT_CONFIG['default'].keys()

    # Public Methods #

    def getDefault(self, key=None):
        config = self._getConfig()
        if key == None:
            return config['default']
        else:
            Validation.validate_string(key, 'getDefault', min_len=1)
            return config['default'][key]

    def getSortersList(self) -> List:
        return self._getConfig()['sorters']

    def getAliasesList(self) -> List:
        return self._getConfig()['aliases']

    def print_defaults(self):
        blacklisted_keys = ['api_key']
        sprint(Styler.stylize('Default:', color='main'))
        for key, value in self.getDefault().items():
            if key in blacklisted_keys:
                continue
            key = key.replace('_', ' ').title()
            sprint(Styler.stylize(
                f'     {key}:  {"N/A" if str(value) == "" else value}', color='main'))

    def print_sorters(self):
        for i, [name, docstr, _] in enumerate(self.getSortersList()):
            sprint(
                Styler.stylize(
                    f'Sorter #{i + 1}, "{name}":  {f"{docstr[:300 - 3]}..." if len(docstr) > 300 else docstr}', color='main')
            )

    def print_aliases(self):
        for i, [alias_name, path] in enumerate(self.getAliasesList()):
            sprint(
                Styler.stylize(
                    f'Alias #{i + 1}, "{alias_name}":  {path}', color='main'
                )
            )



from typing import Dict
import os

from civitdl.api.sorter import SorterData, DirName


def sort_model(model_dict: Dict, version_dict: Dict, filename: str, root_path: str):
    """This is useful when you only just want all of the lora/checkpoint/etc. in the same root directory. Metadata + images are downloaded inside a directory under the parent directory."""
    model_dir_name = DirName.replace_with_rule_1(model_dict['name'])
    model_dir_path = os.path.join(root_path, model_dir_name)
    extra_data_dir_path = os.path.join(
        model_dir_path, f'extra_data-vid_{version_dict["id"]}')

    return SorterData(
        model_dir_path=model_dir_path,
        metadata_dir_path=extra_data_dir_path,
        image_dir_path=extra_data_dir_path,
        prompt_dir_path=extra_data_dir_path
    )




import argparse
import os


from civitconfig.data.configmanager import ConfigManager
from helpers.argparse import PwdAction, ColoredArgParser, BooleanOptionalAction
from helpers.core.utils import get_version


def parse_sorter(sorters, sorter_str):
    for sorter in sorters:
        if sorter_str == sorter[0]:
            return sorter[2]

    return sorter_str


def parse_rootdir(aliases, path):
    dirs = path.split(os.path.sep)

    for alias in aliases:
        if alias[0] == dirs[0]:
            newpath = os.path.join(alias[1], *dirs[1:])
            return parse_rootdir(aliases, newpath)

    return path


HELP_MESSAGE_FOR_SRC_MODEL = """
Provide one or more sources as arguments.
A source can be one of the following: model ID, CivitAI URL, string list of sources or batchfiles.
- batchfiles must be a textfile of comma separated list of sources.
- string list of sources in the following example is "sourceA, sourceB, sourceC, ...":
- Example 1: civitdl source1 "sourceA, sourceB, sourceC, ..." source2 ./path/to/root/model/directory
- Example 2: civitdl ./batchfile1.txt 123456 ~/Downloads/ComfyUI/models/loras
"""

parser = ColoredArgParser(
    prog='civitdl',
    description="A CLI python script to batch download models from CivitAI with CivitAI Api V1.",
    formatter_class=argparse.RawTextHelpFormatter
)
parser.add_argument('srcmodels', type=str, action="extend", nargs='+',
                    help=HELP_MESSAGE_FOR_SRC_MODEL)
parser.add_argument('rootdir', type=str,
                    help='Root directory of where the downloaded model should go.')


parser.add_argument('-s', '--sorter', type=str,
                    help='Specify which sorter function to use.\nDefault is "basic" sorter.\nProvide file path to sorter function if you wish to use a custom sorter.')
parser.add_argument('-i', '--max-images', metavar='INT', type=int,
                    help='Specify max images to download for each model.')

parser.add_argument(
    '--nsfw-mode', metavar='MODE', type=str, help='Specify the nsfw mode when downloading images. Setting to 0 means the program will always download sfw. Setting to 1 means the program will only download sfw + nsfw images if the model itself is nsfw. Setting to 2 means the program will always download both sfw and nsfw.'
)

parser.add_argument('-k', '--api-key', action=PwdAction, type=str, required=False, nargs='?',
                    help='Prompt user for api key to download models that require users to log in.')

parser.add_argument(
    '--with-prompt', action=BooleanOptionalAction, help='Download images with prompt.'
)

parser.add_argument(
    '--without-model', action=BooleanOptionalAction, help='Download only extra details like metadata, images and hashes.'
)

parser.add_argument(
    '--limit-rate', metavar='BYTE', type=str, help='Limit the download speed/rate of resources downloaded from CivitAI.'
)

parser.add_argument(
    '--retry-count', metavar='INT', type=int, help='Specify max number of times to retry downloading a model if it fails.'
)

parser.add_argument(
    '--pause-time', metavar='FLOAT', type=float, help='Specify the number of seconds to pause between each model\'s download.'
)

parser.add_argument(
    '--cache-mode', metavar='MODE', type=str, help='Specify the cache mode. 0 to not use cache. 1 to use cache and copy existant models based on file path. 2 to use cache and copy existant models based on file path + hashes of model. Note that mode 2 has not been implemented yet. See documentation on github for more info.'
)

parser.add_argument(
    '--model-overwrite', action=BooleanOptionalAction, help='Determine whether to overwrite or skip model download if model is already in path. model=overwrite to overwrite model. no-model-overwrite to skip model.'
)

parser.add_argument(
    '--with-color', action=BooleanOptionalAction, help='Enable styles like colors, background colors and bold/italized texts.'
)

parser.add_argument(
    '--verbose', action=BooleanOptionalAction, help='Prints out traceback and other useful information.'
)

parser.add_argument(
    '-v', '--version', action='version', version=f'civitdl v{get_version()}', help='Prints out the version of the program.'
)


def get_args():
    parser_result = parser.parse_args()
    config_manager = ConfigManager()
    config_defaults = config_manager.getDefault()
    sorters = config_manager.getSortersList()
    aliases = config_manager.getAliasesList()

    return {
        "source_strings": parser_result.srcmodels,
        "rootdir": parse_rootdir(aliases, parser_result.rootdir),

        "sorter": parse_sorter(sorters, parser_result.sorter or config_defaults.get('sorter', None)),
        "max_images": parser_result.max_images or config_defaults.get('max_images', None),
        "nsfw_mode": parser_result.nsfw_mode or config_defaults.get('nsfw_mode', None),
        "api_key": parser_result.api_key or config_defaults.get('api_key', None),

        "with_prompt": parser_result.with_prompt if parser_result.with_prompt is not None else config_defaults.get('with_prompt', None),
        "without_model": parser_result.without_model if parser_result.without_model is not None else config_defaults.get('without_model', None),
        "limit_rate": parser_result.limit_rate or config_defaults.get('limit_rate', None),
        "retry_count": parser_result.retry_count or config_defaults.get('retry_count', None),
        "pause_time": parser_result.pause_time or config_defaults.get('pause_time', None),

        "cache_mode": parser_result.cache_mode or config_defaults.get('cache_mode', None),
        "model_overwrite": parser_result.model_overwrite if parser_result.model_overwrite is not None else config_defaults.get('model_overwrite', None),

        "with_color": parser_result.with_color if parser_result.with_color is not None else config_defaults.get('with_color', None),

        "verbose": False if parser_result.verbose == None else parser_result.verbose
    }




import os
import math
import csv
from typing import Dict, Union

from helpers.core.utils import Styler, print_verbose, sprint
from helpers.core.constants import app_dirs

# TODO: Watch out for edge cases where one of the hash is empty.
# { '123456': { 'model_filepath': 'path', 'SHA256': 'hash1', 'BLAKE3': 'hash2' } }


class HashManager:
    __CACHE_COLUMNS = ['volume_id', 'model_filepath', 'SHA256', 'BLAKE3']
    __version_id: str
    __filepath: str
    __hashes_dict: Dict

    def __init__(self, version_id: str):
        self.__version_id = version_id
        version_id_num = int(version_id)

        dirpath = os.path.join(app_dirs.user_cache_dir,
                               'hashes', str(math.floor(version_id_num / 10000)))
        print_verbose(dirpath)
        filenum = math.floor(version_id_num / 100)
        filename = f'{filenum}.csv'
        self.__filepath = os.path.join(dirpath, filename)
        if not os.path.isfile(self.__filepath):
            os.makedirs(os.path.dirname(self.__filepath), exist_ok=True)
            with open(self.__filepath, 'w', encoding='UTF-8') as file:
                csv_writer = csv.writer(file)
                csv_writer.writerow(self.__CACHE_COLUMNS)

        self.__hashes_dict = self.__read_from_csv()

    def __read_from_csv(self):
        hashes_dict = {}
        with open(self.__filepath, 'r', encoding='UTF-8') as f:
            csv_reader = csv.reader(f)
            next(csv_reader)
            for row in csv_reader:
                if row == []:
                    continue
                hashes_dict[row[0]] = {
                    'model_filepath': row[1],
                    'SHA256': row[2],
                    'BLAKE3': row[3]
                }
        return hashes_dict

    def __write_to_csv(self):
        with open(self.__filepath, 'w', encoding='UTF-8') as f:
            csv_writer = csv.writer(f)
            csv_writer.writerow(
                self.__CACHE_COLUMNS)
            for key, value in self.__hashes_dict.items():
                csv_writer.writerow([
                    key,
                    value.get('model_filepath', ''),
                    value.get('SHA256', ''),
                    value.get('BLAKE3', '')
                ])

    def __get_hash_dict(self) -> Dict:
        return self.__hashes_dict.get(self.__version_id)

    def set_local_model_cache(self, model_filepath: str, hashes: Dict[str, str]) -> None:
        self.__hashes_dict[self.__version_id] = {
            'model_filepath': os.path.abspath(model_filepath),
            'SHA256': hashes.get('SHA256', ''),
            'BLAKE3': hashes.get('BLAKE3', '')
        }
        self.__write_to_csv()

    def get_local_model_path(self) -> Union[None, str]:
        hash_dict = self.__get_hash_dict()

        if hash_dict is None:
            sprint(Styler.stylize(
                f'Cache of model with version id {self.__version_id} not found.', color='info'))
            return None
        else:
            filepath = hash_dict['model_filepath']
            return None if not os.path.isfile(filepath) else filepath



from dataclasses import dataclass
import os
import re
from typing import List, Literal, Union

from helpers.core.utils import print_verbose, InputException, UnexpectedException


@dataclass
class Id:
    type: Literal['id', 'site', 'api']
    data: List[str]
    original: str

    def __init__(self, type, data: List[str], original: str):
        if type != 'id' and type != 'site' and type != 'api':
            raise UnexpectedException(
                f'Unknown type provided for Id class: {type}')

        self.type = type
        self.data = data
        self.original = original

    def __post_init__(self):
        if self.type != 'id' and self.type != 'site' and self.type != 'api':
            raise UnexpectedException(
                f'Unknown type provided for Id class: {self.type}')
        for el in self.data:
            if not isinstance(el, str):
                raise UnexpectedException(
                    f'Wrong data type for {el} in {self.data}')

        if not isinstance(self.original, str):
            raise UnexpectedException(f'Wrong data type for {self.original}')


class SourceManager:
    def __init__(self) -> None:
        pass

    def __get_comma_list(self, string: str) -> List[str]:
        return [input_str for input_str in string.replace(
            '\n', '').split(',') if input_str.strip() != '']

    def __use_parent_dir_if_exist(self, src: str, parent: Union[str, None]) -> str:
        return os.path.normpath(os.path.join(os.path.dirname(parent), src)) if parent else src

    def parse_src(self, str_li: List[str], parent: Union[str, None] = None) -> List[Id]:
        res: List[Id] = []
        for string in str_li:
            string = string.strip()

            if string.isdigit() and abs(int(string)) == int(string):
                res.append(Id('id', [string], string))
            elif len(self.__get_comma_list(string)) > 1:
                arg_str_li = self.__get_comma_list(string)
                res.extend(self.parse_src(arg_str_li))
            elif 'civitai.com/api' in string:
                version_id_regex = r'(?<=models\/)\d+'
                version_id = re.search(version_id_regex, string)
                if version_id == None:
                    err = 'Incorrect format for the url provided' + \
                        (f' in {parent}: ' if parent else ': ') + string
                    raise InputException(err)
                version_id = version_id.group(0)
                res.append(Id('api', [version_id], string))
            elif 'civitai.com/models' in string:
                model_id = re.search(r'(?<=models\/)\d+', string)
                version_id = re.search(r'(?<=modelVersionId=)\d+', string)

                if model_id == None:
                    err = 'Incorrect format for the url/id provided' + \
                        (f' in {parent}: ' if parent else ': ') + string
                    raise InputException(err)
                model_id = model_id.group(0)

                if version_id:
                    version_id = version_id.group(0)
                    res.append(
                        Id('site', [model_id, version_id], string))
                else:
                    res.append(Id('site', [model_id], string))
            elif os.path.exists(self.__use_parent_dir_if_exist(string, parent)):
                string = self.__use_parent_dir_if_exist(string, parent)
                file_str = None

                with open(string, 'r', encoding='UTF-8') as file:
                    file_str = file.read().strip()
                if file_str == None:
                    raise UnexpectedException(
                        'Unknown exception while reading batchfile path.')
                str_li_res = self.__get_comma_list(file_str)
                print_verbose(str_li_res)
                res.extend(self.parse_src(str_li_res, parent=string))
            else:
                raise InputException(
                    f'Bad source provided: {string}', f'   Parent Batchfile Path: {parent}' if parent else None)

        return res



from enum import Enum
from typing import List, Literal

# Private


class _FormatEnum(Enum):
    @classmethod
    def get_attribute(cls, format: str):
        format = format.upper()
        for attr in cls:
            if format == attr.name:
                return attr


class _Style(_FormatEnum):
    # Text styles
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class _Colors(_FormatEnum):
    # Text colors
    EXCEPTION = '\033[31m'
    WARNING = '\033[33m'
    SUCCESS = '\033[32m'

    PRIMARY = '\033[34m'
    SECONDARY = '\033[35m'
    INFO = '\033[36m'


class _BGColors(_FormatEnum):
    # Background colors
    EXCEPTION = '\033[41m'
    WARNING = '\033[43m'
    SUCCESS = '\033[42m'

    PRIMARY = '\033[44m'
    SECONDARY = '\033[45m'
    INFO = '\033[46m'


def _color_generator():
    color = 'primary'
    while True:
        if color == 'primary':
            color = 'secondary'
            yield 'primary'
        else:
            color = 'primary'
            yield 'secondary'


_color_iter = _color_generator()
_bgcolor_iter = _color_generator()

_no_style = False

# Public / Exports


def disable_style():
    global _no_style
    _no_style = True


class Styler(Enum):
    RESET = '\033[0m'  # Reset to default text and background color

    @classmethod
    def _get_main_color_attr(cls, color: str):
        color = color.upper()
        if color == 'MAIN':
            color = cls._get_next_color()
        elif color in cls._get_main_color_list():
            raise Exception(
                f'Can not directly access PRIMARY or SECONDARY color.')

        return _Colors.get_attribute(color)

    @classmethod
    def _get_main_bgcolor_attr(cls, bgcolor: str):
        bgcolor = bgcolor.upper()
        if bgcolor == 'MAIN':
            bgcolor = cls._get_next_bgcolor()
        elif bgcolor in cls._get_main_color_list():
            raise Exception(
                f'Can not directly access PRIMARY or SECONDARY background color.')

        return _BGColors.get_attribute(bgcolor)

    @staticmethod
    def _get_next_color():
        return next(_color_iter)

    @staticmethod
    def _get_next_bgcolor():
        return next(_bgcolor_iter)

    @staticmethod
    def _get_main_color_list():
        return ['PRIMARY', 'SECONDARY']

    @classmethod
    def stylize(cls, string: str, color: str = None, bg_color: str = None, styles: List[str] = None):
        res = ''

        if not isinstance(string, str):
            raise Exception(f'String provided to stylize is not a string')

        if _no_style:
            return str(string)

        if styles:
            for style in styles:
                style_attr = _Style.get_attribute(style)
                if style_attr:
                    res += style_attr.value
                else:
                    raise Exception(
                        f'The following style "{style}" does not exist.')

        if color:
            color_attr = cls._get_main_color_attr(color)
            if color_attr:
                res += color_attr.value
            else:
                raise Exception(
                    f'The following color "{color}" does not exist.')

        if bg_color:
            bg_attr = cls._get_main_bgcolor_attr(bg_color)
            if bg_attr:
                res += bg_attr.value
            else:
                raise Exception(
                    f'The following background color "{bg_color}" does not exist.')

        res += string
        res += cls.RESET.value
        return str(res)


class CustomException(Exception):
    def __init__(self, error_type, *messages):
        res = ""

        res += Styler.stylize(f'\n{self.__class__.__name__} ({error_type}):',
                              color='exception', styles=['bold'])

        for message in messages:
            res += Styler.stylize(f'\n         {message}', color='exception')

        super().__init__(res)


class InputException(CustomException):
    def __init__(self, *messages):
        super().__init__('Bad Inputs', *messages)


class ResourcesException(CustomException):
    def __init__(self, *messages):
        super().__init__('Resources Not Found', *messages)


class APIException(CustomException):
    def __init__(self, status_code, *messages):
        super().__init__(f'API Status Code {status_code}', *messages)


class NotImplementedException(CustomException):
    def __init__(self, *messages):
        super().__init__('Feature Not Implemented', *messages)


class UnexpectedException(CustomException):
    def __init__(self, *messages):
        super().__init__('Unexpected Error', *messages)
